from __future__ import annotations

import io
import os
import zipfile
from datetime import datetime, timezone
from typing import TypedDict

from flask import Blueprint, Response, send_file
from flask.views import MethodView

import ckan.plugins.toolkit as tk

from ckanext.tables.shared import GenericTableView

from ckanext.logs.config import get_logs_folder
from ckanext.logs.table import LogsTable

bp = Blueprint("logs", __name__, url_prefix="/ckan-admin/logs")


class LogFileInfo(TypedDict):
    name: str
    size: int
    mtime: datetime


def before_request() -> None:
    try:
        tk.check_access("sysadmin", {"user": tk.current_user.name})
    except tk.NotAuthorized:
        tk.abort(403, tk._("Need to be system administrator to administer"))


class LogsDashboardView(MethodView):
    def get(self):
        logs_folder = get_logs_folder()
        all_log_files = self.get_logs_files(logs_folder)
        total_size = sum(f["size"] for f in all_log_files)
        logs_files = [f for f in all_log_files if f["name"].endswith(".log")]
        logs_sizes = self._calculate_logs_sizes(all_log_files, logs_files)

        return tk.render(
            "logs/dashboard.html",
            {
                "logs_folder": logs_folder,
                "logs_files": logs_files,
                "total_size": total_size,
                "logs_sizes": logs_sizes,
            },
        )

    def get_logs_files(self, logs_folder: str | None) -> list[LogFileInfo]:
        if not logs_folder or not os.path.exists(logs_folder):
            return []

        files = []

        for name in os.listdir(logs_folder):
            path = os.path.join(logs_folder, name)
            if os.path.isfile(path):
                size = os.path.getsize(path)
                mtime = os.path.getmtime(path)
                files.append(
                    LogFileInfo(
                        name=name,
                        size=size,
                        mtime=datetime.fromtimestamp(mtime, tz=timezone.utc),
                    )
                )

        return sorted(files, key=lambda x: x["mtime"], reverse=True)

    def _calculate_logs_sizes(self, all_files: list[LogFileInfo], log_files: list[LogFileInfo]) -> dict[str, int]:
        result = {}

        for log_file in log_files:
            base_name = log_file["name"].strip(".log")
            total_size = log_file["size"]

            for file in all_files:
                if file["name"].startswith(base_name) and file["name"] != log_file["name"]:
                    total_size += file["size"]

            result[log_file["name"]] = total_size

        return result


class LogsDownloadView(MethodView):
    def get(self) -> Response:
        logs_folder = get_logs_folder()

        if not logs_folder or not os.path.exists(logs_folder):
            tk.abort(404, "Logs directory not found")

        zip_buffer = io.BytesIO()

        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zf:
            for name in os.listdir(logs_folder):
                path = os.path.join(logs_folder, name)
                if os.path.isfile(path):
                    zf.write(path, arcname=name)

        zip_buffer.seek(0)

        return send_file(zip_buffer, as_attachment=True, download_name="logs_archive.zip", mimetype="application/zip")


class LogsExportTableView(MethodView):
    def get(self, log_file: str) -> Response:
        logs_folder = get_logs_folder()
        log_base_name = log_file.strip(".log")

        if not logs_folder or not os.path.exists(logs_folder):
            tk.abort(404, "Logs directory not found")

        zip_buffer = io.BytesIO()

        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zf:
            for name in os.listdir(logs_folder):
                if not name.startswith(log_base_name):
                    continue

                path = os.path.join(logs_folder, name)

                if os.path.isfile(path):
                    zf.write(path, arcname=name)

        zip_buffer.seek(0)

        return send_file(zip_buffer, as_attachment=True, download_name="logs_archive.zip", mimetype="application/zip")


class LogsGenericTableView(GenericTableView):
    def get(self, log_file: str) -> str | Response:   # type: ignore
        table = self.table(log_file=log_file)  # type: ignore

        if exporter_name := tk.request.args.get("exporter"):
            return self._export(table, exporter_name)

        if tk.request.headers.get("X-Requested-With") == "XMLHttpRequest":
            return self._ajax_data(table)

        return table.render_table(
            breadcrumb_label=self.breadcrumb_label,
            page_title=self.page_title,
        )

bp.before_request(before_request)

bp.add_url_rule("/dashboard", view_func=LogsDashboardView.as_view("dashboard"))
bp.add_url_rule("/export", view_func=LogsDownloadView.as_view("download"))
bp.add_url_rule("/export_log_file/<log_file>", view_func=LogsExportTableView.as_view("export_log_file"))
bp.add_url_rule("/table/<log_file>", view_func=LogsGenericTableView.as_view("table", table=LogsTable))



from __future__ import annotations

import gzip
import re
from datetime import datetime
from pathlib import Path
from typing import Any

from ckanext.tables.shared import ListDataSource

from ckanext.logs import config

LOG_ENTRY_START_RE = re.compile(
    r"^(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3})\s+"
    r"(INFO|ERROR|WARNING|WARNI|DEBUG|CRITICAL)\s+"
    r"\[([^\]]+)\]\s*(.*)"
)

UWSGI_LOG_PATTERN = re.compile(
    r"""
    ^\[pid:                             # must start like a uwsgi log
    .*?                                 # skip until timestamp
    \[                                  # opening bracket before timestamp
    (?P<timestamp>                      # capture the timestamp
        [A-Za-z]{3}\s+                  # day of week, e.g. Sun
        [A-Za-z]{3}\s+                  # month, e.g. Dec
        \d{1,2}\s+                      # day of month
        \d{2}:\d{2}:\d{2}\s+            # time
        \d{4}                           # year
    )
    \]                                  # closing bracket
""",
    re.VERBOSE,
)


class LogDataSource(ListDataSource):
    """Data source for reading log files with multi-line entries."""

    def __init__(self, log_file: str, n_logs: int = 10000):
        self.logs_path = Path(config.get_logs_folder() or "")
        self.logs_filename = log_file
        self.n_logs = n_logs
        self.data: list[dict[str, Any]] = []
        self.filtered: list[dict[str, Any]] | None = None

        self._load_all_logs()

    def _load_all_logs(self):
        """Load all log entries from all files, newest first."""
        log_files = sorted(
            self.logs_path.glob(f"{self.logs_filename}*"),
            key=lambda p: p.stat().st_mtime,
            reverse=True,
        )

        all_entries: list[dict[str, Any]] = []

        for log_file in log_files:
            entries = self._parse_lines(self._read_all_lines(log_file))
            all_entries.extend(entries)

            if len(all_entries) >= self.n_logs:
                break

        # sort by timestamp descending
        all_entries.sort(key=lambda e: e.get("timestamp", ""), reverse=True)
        self.data = all_entries
        self.filtered = self.data

    def _read_all_lines(self, log_file: Path) -> list[str]:
        """Read all lines from a file (supports .gz)."""
        try:
            if log_file.suffix == ".gz":
                with gzip.open(log_file, "rt", errors="ignore") as f:
                    return f.readlines()
            else:
                with log_file.open("r", errors="ignore") as f:
                    return f.readlines()
        except Exception:  # noqa
            return []

    def _parse_lines(self, lines: list[str]) -> list[dict[str, Any]]:
        entries = []
        buffer = []  # buffer for multi-line messages
        current_entry = None

        for line in lines:
            if match := UWSGI_LOG_PATTERN.match(line):
                # Flush previous entry
                if current_entry:
                    if buffer:
                        current_entry["message"] += "\n" + "\n".join(buffer)
                    entries.append(current_entry)
                    buffer = []

                current_entry = {
                    "timestamp": self.parse_uwsgi_timestamp(match.group("timestamp")),
                    "level": "UWSGI",
                    "module": "uwsgi",
                    "message": line.strip(),
                }
            elif match := LOG_ENTRY_START_RE.match(line):
                # Flush previous entry
                if current_entry:
                    if buffer:
                        current_entry["message"] += "\n" + "\n".join(buffer)
                    entries.append(current_entry)
                    buffer = []

                timestamp, level, module, message = match.groups()
                current_entry = {
                    "timestamp": timestamp,
                    "level": level,
                    "module": module,
                    "message": message.strip(),
                }
            else:
                buffer.append(line.rstrip())

        # Flush last entry
        if current_entry:
            if buffer:
                current_entry["message"] += "\n" + "\n".join(buffer)
            entries.append(current_entry)

        return entries

    def parse_uwsgi_timestamp(self, timestamp: str) -> str | None:
        dt = datetime.strptime(timestamp, "%a %b %d %H:%M:%S %Y")  # noqa: DTZ007

        # to match the format of other log timestamps
        return dt.strftime("%Y-%m-%d %H:%M:%S,%f")[:-3]



import math


def logs_printable_file_size(size_bytes: int) -> str:
    if size_bytes == 0:
        return "0 bytes"

    size_name = ("bytes", "KB", "MB", "GB", "TB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(float(size_bytes) / p, 1)

    return f"{s} {size_name[i]}"



import ckan.plugins.toolkit as tk

CONF_LOGS_FOLDER = "ckanext.logs.logs_path"
CONF_LOGS_FILE_NAME = "ckanext.logs.log_filename"


def get_logs_folder() -> str | None:
    return tk.config[CONF_LOGS_FOLDER]



import ckan.plugins.toolkit as tk

from ckanext.tables import types
from ckanext.tables.formatters import BaseFormatter, DialogModalFormatter


class LogsDialogModalFormatter(BaseFormatter):
    """Formatter to show log details in a modal dialog."""

    def format(self, value: types.Value, options: types.Options) -> types.FormatterResult:
        formatter = DialogModalFormatter(self.column, self.row, self.initial_row, self.table)
        value = self.initial_row["message"]

        options.update({"template": "logs/formatters/dialog_modal.html"})

        return formatter.format(value, options)



from __future__ import annotations

import ckan.plugins.toolkit as tk
from ckan.types import Context

from ckanext.tables.formatters import TrimStringFormatter
from ckanext.tables.shared import ActionHandlerResult, ColumnDefinition, TableActionDefinition, TableDefinition

from ckanext.logs.data_source import LogDataSource
from ckanext.logs.formatters import LogsDialogModalFormatter


class LogsTable(TableDefinition):
    """Table definition for the logs dashboard."""

    def __init__(self, log_file: str):
        """Initialize the table definition."""
        self.log_file = log_file

        super().__init__(
            name="logs",
            data_source=LogDataSource(log_file=log_file),
            columns=[
                ColumnDefinition(field="timestamp", width=180, resizable=False),
                ColumnDefinition(field="level", width=100, resizable=False),
                ColumnDefinition(field="module", width=200),
                ColumnDefinition(
                    field="message",
                    formatters=[(TrimStringFormatter, {"max_length": 88})],
                    tabulator_formatter="html",
                ),
                ColumnDefinition(
                    title=" ",
                    field="details",
                    formatters=[(LogsDialogModalFormatter, {"max_length": 88})],
                    tabulator_formatter="html",
                    width=50,
                    sortable=False,
                    resizable=False,
                    filterable=False,
                ),
            ],
            table_actions=[
                TableActionDefinition(
                    action="export_logs",
                    label=tk._("Export Logs"),
                    icon="fa fa-download",
                    callback=self.table_action_export_logs,
                ),
            ]
        )

    def table_action_export_logs(self) -> ActionHandlerResult:
        return ActionHandlerResult(
            success=True,
            error=None,
            redirect=tk.h.url_for("logs.export_log_file", log_file=self.log_file),
        )

    @classmethod
    def check_access(cls, context: Context) -> None:
        tk.check_access("sysadmin", context)



# -*- coding: utf-8 -*-
from setuptools import setup

# Note: Do not add new arguments to setup(), instead add setuptools
# configuration options to setup.cfg, or any other project information
# to pyproject.toml
# See https://github.com/ckan/ckan/issues/8382 for details

setup(
    # If you are changing from the default layout of your extension, you may
    # have to change the message extractors, you can read more about babel
    # message extraction at
    # http://babel.pocoo.org/docs/messages/#extraction-method-mapping-and-configuration
    message_extractors={
        'ckanext': [
            ('**.py', 'python', None),
            ('**.js', 'javascript', None),
            ('**/templates/**.html', 'ckan', None),
        ],
    }
)
